#lang racket

(require redex (except-in "main.rkt" F lookup ∪⊗) "util.rkt")

(provide (all-defined-out))

(define-extended-language cps 
  anf
  [Cp (fn Ck Ce)]
  [Ce (Ct Ct Cc) (Cc Ct)]
  [Ct Cx Cv (fn Cx Ck Ce) c]
  ;; continuation variables must start with k
  [Ck (side-condition (name n variable-not-otherwise-mentioned)
                         (eq? #\k (string-ref (symbol->string (term n)) 0)))]
  ;; intermediate variables (or really program variables?)
  [Cv (name n variable-not-otherwise-mentioned)]
  [Cx variable-not-otherwise-mentioned]
  [Cc (fn Cv Ce) Ck]
  [CVar Cx Cv Ck]
  [CProg Cp]
  [SExp Ce]
  [TExp Ct]
  [CExp Cc]
  [Env# ([CVar (Val# ...)] ...)]
  [R# Env#]
  [(w# Val#) stop (fn Cx Ck Ce) (fn Cv Ce) c]
  [Vals#/f (Val# ...) #f]
  ;; trivial w/ no constants
  [nce-t x (fn x nce-s)]
  [nce-s nce-t (let x nce-t nce-s) (nce-t nce-t) (let x (nce-t nce-t) nce-s)])

(define-metafunction cps
  C : P -> CProg
  [(C p) (fn Ck (F Ck p))         
         (where Ck ,(gensym 'k))])

(define-metafunction cps
  F : Ck C -> SExp
  [(F Ck t) (Ck (V t))]
  [(F Ck (t_0 t_1)) ((V t_0) (V t_1) Ck)]
  [(F Ck (let x t s)) ((fn x (F Ck s)) (V t))]
  [(F Ck (let x (t_0 t_1) s)) ((V t_0) (V t_1) (fn x (F Ck s)))])

(define-metafunction cps
  V : T -> TExp
  [(V x) x]
  [(V c) c]
  [(V (fn x s)) (fn x Ck (F Ck s))         
                (where Ck ,(gensym 'k))])

(define-metafunction cps
  D : CProg -> P
  [(D (fn Ck Ce)) (U Ce)])

(define-metafunction cps
  U : SExp -> C
  [(U (Ck Ct)) (P Ct)]
  [(U ((fn Cv Ce) Ct))
   (let Cv (P Ct) (U Ce))]
  [(U (Ct_0 Ct_1 Ck)) ((P Ct_0) (P Ct_1))]
  [(U (Ct_0 Ct_1 (fn Cv Ce))) 
   (let Cv ((P Ct_0) (P Ct_1)) (U Ce))])

(define-metafunction cps
  P : TExp -> T
  [(P Cx) Cx]
  [(P Cv) Cv]
  [(P c) c]
  [(P (fn Cx Ck Ce)) (fn Cx (U Ce))])

(define (l5.1 [n 1000])
  (redex-check cps
               (p s t)
               (let ([k (gensym 'k)])
                 (and (equal? (term (D (C p))) (term p))
                      (equal? (term (U (F ,k s))) (term s))
                      (equal? (term (P (V t))) (term t))))
                #:attempts n))



(define-metafunction cps
  lookup : Var Env# -> Vals#/f
  [(lookup Cx ((Cx any_v) any_r ...)) any_v]
  [(lookup Cx_1 ((Cx_2 any_v) any_r ...)) (lookup Cx_1 (any_r ...))]
  [(lookup Cx ()) #f])

(define-metafunction cps
  μ#t : TExp Env# -> (Val# ...)
  [(μ#t Cx R#) (lookup Cx R#)]
  [(μ#t Cv R#) (lookup Cv R#)]
  [(μ#t (fn Cx Ck Ce) R#) ((fn Cx Ck Ce))]
  [(μ#t c R#) (c)])

(define-metafunction cps
  μ#c : CExp Env# -> (Val# ...)
  [(μ#c Ck R#) (lookup Ck R#)]
  [(μ#c (fn Cv Ce) R#) ((fn Cv Ce))])

(define (T#-init p)
  (match p
    [`(fn ,k ,e)
     (term [(,e) ((kr (stop)) 
                  (,k ((fn vr (kr vr)))))])]))

(define Q#? (redex-match cps (SExp ...)))
(define R#? (redex-match cps R#))

(define (chk-Q#-R# what Q#-R#)
  (match Q#-R#
    [`(,Q# ,R#)
     (unless (Q#? Q#) (error 'Q# "~a has illformed Q#: ~s" what Q#))
     (unless (R#? R#) (error 'R# "~a has illformed R#: ~s" what R#))
     Q#-R#]))

(define (∪⊗ arg1 . args)
  (define all-args (cons arg1 args))
  (for ([x (in-list all-args)]
        [y (in-naturals)])
    (unless (and (list? x)
                 (= (length x) 2))
      (error '∪⊗ 
             "expected all arguments to be lists of length 2, but argument ~a is ~s"
             y
             x)))
  (define fst (map first all-args))
  (define snd (map second all-args))
  (list (apply ∪ fst)
        (term (∪· ,@snd))))

(define ((T# p) Q#-R#)
  (match (chk-Q#-R# 'input Q#-R#)
    [`(,Q# ,R#)
     (chk-Q#-R# 
      'result
      (apply ∪⊗
             (append
              (list (T#-init p))
              (set-of cps
                      ((Ce) (∪· ,R# ((Cx (μ#t Ct_1 ,R#)) (Ck (μ#c Cc ,R#)))))
                      (Ct_0 Ct_1 Cc) ∈ ,Q#
                      (fn Cx Ck Ce) ∈ (μ#t Ct_0 ,R#))
              
              (set-of cps
                      ((Ce) (∪· ,R# ((Cv (μ#t Ct ,R#)))))
                      (Cc Ct) ∈ ,Q#
                      (fn Cv Ce) ∈ (μ#c Cc ,R#)))))]))

;; variables bound to the stop continuation
(define stop-cont? (redex-match cps (Ck [any_1 ... stop any_2 ...])))

(define (T#-envs Q#-R#)
  (match Q#-R#
    [`(,Q# ,R#) R#]))

(define (T#-answers Q#-R#)
  (match Q#-R#
    [`(,Q# ,R#)     
     ;; the variables bound to `stop'
     (define final-conts (map first (filter stop-cont? R#)))
     ;; values/variables that are arguments to those continuations 
     (define return-vals (map second (filter (λ (app) (memq (first app) final-conts))
                                             Q#)))
     (apply append (map (λ (var)
                          (if (symbol? var)                              
                              (map (λ (v) (term (P ,v))) (term (fapp ,R# ,var)))
                              (list var)))
                        return-vals))]))

(define simple (term ((fn x x) 1)))

(define simple-cps (term (C ,simple)))

(define big (term (let g (fn z (let f (fn y 1) (let x 2 (f x))))
                    (let f (fn w 3) (let x (g 1) (f x))))))
(define big-cps (term (C ,big)))

(define (test)
  'starting
  (T#-answers (lfp (T# (term (C ,simple))) '(() ())))
  (F#-answers (lfp (F# simple) '(() () ())))
  
  (T#-answers (lfp (T# (term (C ,big))) '(() ())))
  (F#-answers (lfp (F# big) '(() () ())))
  
  (T#-answers (lfp (T# (term (C (fn g 1)))) '(() ())))
  (F#-answers (lfp (F# (term (fn g 1))) '(() () ()))))
  
(define (analyze t)
  (values 
   (T#-answers (lfp (T# (term (C ,t))) '(() ())))
   (F#-answers (lfp (F# (term ,t)) '(() () ())))))

(define om '(fn x (x x)))
(define omom `(let w ,om (w w)))

(define ex '(let J ((fn t 1)
                    (fn ZX (let F -13 (2 1)))) 
              (let bLt (0 F) (let tF 4 1))))
(define ex2 '(let J ((fn t 1) (fn ZX ZX))
                 (let bLt (0 1) 1)))
(define ex3 '(let W (fn N N)
                 (let J ((fn t W) (fn ZX ZX)) 
                 (let bLt (0 1) W))))
(define ex4 `(let W (fn N (N N))
               (let id (fn XX XX)
                 (let id2 (fn QQ QQ)
                   (let J ((fn t id2) id) 
                     (let bLt (W W) id2))))))
