#lang scheme

(require redex/reduction-semantics
         "grammar.rkt"
         "typing.rkt")

(define (compile e)
  (unless (term (typeof () ,e))
    (error 'compile "ill-typed ~s" e))
  (term (CPS (ANF ,e))))

(define-metafunction SR
  CPS : e -> e
  [(CPS e) 
   e_^’
   (where (e_^’) (CPS’ e ()))])

; Figure 4 with the environment as an extra parameter
(define-metafunction SR
  CPS’ : r ... Γ -> (r ...)
  
  [(CPS’ (shift e) Γ)
   (CPS’ e Γ)]
  
  [(CPS’ (reset e) Γ)
   ((e_^’ (block (def id (x : (A ε)) : (B ε) = x) id)))
   (where (A (@cps B C)) (typeof Γ e))
   (where (e_^’) (CPS’ e Γ) )]
  [(CPS’ (reset e) Γ)
   (CPS’ e Γ)
   (where (A ε) (typeof Γ e))]
  
  [(CPS’ (e_1 e_2) Γ)
   ((e_1^’ e_2^’))
   (where (e_1^’) (CPS’ e_1 Γ))
   (where (e_2^’) (CPS’ e_2 Γ))]
  
  [(CPS’ x Γ) (x)]
  [(CPS’ o Γ) (o)]
  [(CPS’ number Γ) (number)]
  [(CPS’ string Γ) (string)]
  
  [(CPS’ (block r ...) Γ)
   ((block r_^’ ...))
   (where (r_^’ ...) (CPS’ r ... Γ))]
  
  [(CPS’ (val x : τ = e) r ... (name Γ ((x_0 τ_0) ...)))
   ((def f (y : ((⇒ (U_^’ ε) (V_^’ ε)) ε)) : (C_^’ ε) =
      (e_^’ (block (def z (x : (A_^’ ε)) : (W_^’ ε) = ((block r_^’ ...) y)) z)))
    f)
   (where (A (@cps B C)) (typeof Γ e))
   (where Γ_^’ ((x (A ε)) (x_0 τ_0) ...))
   (where (U (@cps V W)) (typeof Γ_^’ (block r ...)))
   (where (e_^’) (CPS’ e Γ))
   (where (r_^’ ...) (CPS’ r ... Γ_^’))
   (where (f y z) ,(variables-not-in (term (e_^’ r_^’ ...)) '(f k g)))
   (where (A_^’ C_^’ U_^’ V_^’ W_^’)
          ((CPS-pure-type A)
           (CPS-pure-type C)
           (CPS-pure-type U)
           (CPS-pure-type V)
           (CPS-pure-type W)))]
  [(CPS’ (val x : τ = e) r ... (name Γ ((x_0 τ_0) ...)))
   ((def f (y : ((⇒ (U_^’ ε) (B_^’ ε)) ε)) : (C_^’ ε) =
      (e_^’ (block (def z (x : (A_^’ ε)) : (B_^’ ε) = (y (block r_^’ ...))) z)))
    f)
   (where (A (@cps B C)) (typeof Γ e))
   (where Γ_^’ ((x (A ε)) (x_0 τ_0) ...))
   (where (U ε) (typeof Γ_^’ (block r ...)))
   (where (e_^’) (CPS’ e Γ))
   (where (r_^’ ...) (CPS’ r ... Γ_^’))
   (where (f y z) ,(variables-not-in (term (e_^’ r_^’ ...)) '(f k g)))
   (where (A_^’ B_^’ C_^’ U_^’)
          ((CPS-pure-type A)
           (CPS-pure-type B)
           (CPS-pure-type C)
           (CPS-pure-type U)))]
  [(CPS’ (val x : τ = e) r ... (name Γ ((x_0 τ_0) ...)))
   ((val x : τ_^’ = e_^’) r_^’ ...)
   (where (A ε) (typeof Γ e))
   (where (e_^’) (CPS’ e Γ))
   (where (r_^’ ...) 
          (CPS’ r ... ((x (A ε)) (x_0 τ_0) ...)))
   (where τ_^’ (CPS-type τ))]
  
  [(CPS’ (def f (x : τ_1) : τ_2 = e) r ... ((x_0 τ_0) ...))
   ((def f (x : τ_1^’) : τ_2^’ = e_^’)
    r_^’ ...)
   (where (e_^’) 
          (CPS’ e ((x τ_1) (f ((⇒ τ_1 τ_2) ε)) (x_0 τ_0) ...)))
   (where (r_^’ ...) 
          (CPS’ r ... ((f ((⇒ τ_1 τ_2) ε)) (x_0 τ_0) ...)))
   (where (τ_1^’ τ_2^’) ((CPS-type τ_1) (CPS-type τ_2)))]
  
  [(CPS’ e r ... Γ)
   (e_^’ r_^’ ...)
   (where (e_^’) (CPS’ e Γ))
   (where (r_^’ ...) (CPS’ r ... Γ))])

(define-metafunction SR
  CPS-type : τ -> τ
  [(CPS-type (A ε)) 
   ((CPS-pure-type A) ε)]
  [(CPS-type (A (@cps B C)))
   ((⇒ ((⇒ ((CPS-pure-type A) ε) ((CPS-pure-type B) ε)) ε)
       ((CPS-pure-type C) ε))
    ε)])

(define-metafunction SR
  CPS-pure-type : A -> A
  [(CPS-pure-type Int) Int]
  [(CPS-pure-type Str) Str]
  [(CPS-pure-type (⇒ τ_1 τ_2))
   (⇒ (CPS-type τ_1) (CPS-type τ_2))])

(define-extended-language SR-compile SR
  (g number))

(define-metafunction SR-compile
  first-fresh : r -> g
  [(first-fresh any)
   ,(match (variable-not-in (term any) 'x)
      [(app symbol->string (pregexp #px"^x(\\d+)$" (list _ n)))
       (add1 (string->number n))]
      [_ 0])])

(define-metafunction SR-compile
  next-fresh : g -> (x g)
  [(next-fresh g)
   (,(string->symbol (format "x~s" (term g))) 
    ,(add1 (term g)))])

(define-metafunction SR-compile
  ANF : e -> e
  [(ANF e)
   (block-if-nec r ... e_^’)
   (where ((r ... e_^’) g) (ANF-tail e () (first-fresh e)))])

; [|.|] from figure 5 with two changes:
; 1. An extra parameter, the environment
; 2. State threaded through the transformation to supply fresh variables. 
; (It's hard to tell what to choose as the first argument to `variable-not-in'.)
(define-metafunction SR-compile
  ANF-tail : e Γ g -> ((r ...) g)
  [(ANF-tail (e_1 e_2) Γ g)
   ((r_0 ... r_i ... (e_1^’ e_2^’)) g_^’’)
   (where ((r_0 ... e_1^’) g_^’) (ANF-ctxt e_1 Γ g))
   (where ((r_i ... e_2^’) g_^’’) (ANF-ctxt e_2 Γ g_^’))]
  [(ANF-tail (block r ... e) Γ g)
   ; Interpret resulting block as a singleton sequence.
   (((block s_0 ... s_i ... e_^’)) g_^’’)
   (where ((s_0 ...) g_^’) (ANF-ctxt r ... Γ g))
   (where ((s_i ... e_^’) g_^’’) 
          ; Don't need the inserted bindings from r_0 ...
          (ANF-tail e (extend Γ (r ...)) g_^’))]
  [(ANF-tail (shift e) Γ g)
   (((shift e_^’)) g_^’)
   ; e is pure so nothing will be lifted
   (where ((e_^’) g_^’) (ANF-tail e Γ g))]
  [(ANF-tail (reset e) Γ g)
   (((reset (block-if-nec r ... e_^’))) g_^’)
   (where ((r ... e_^’) g_^’) (ANF-tail e Γ g))]
  [(ANF-tail o Γ g) ((o) g)]
  [(ANF-tail string Γ g) ((string) g)]
  [(ANF-tail number Γ g) ((number) g)]
  [(ANF-tail x Γ g) ((x) g)])

(define-metafunction SR-compile
  ANF-ctxt : r ... Γ g -> ((r ...) g)
  [(ANF-ctxt e ((x_0 τ_0) ...) g)
   ((r_i ... (val x_i+1 : (A ε) = e_^’) x_i+1) g_^’’)
   (where (((name r_i (val x_i : τ_i = e_i)) ... e_^’) g_^’)
          (ANF-tail e ((x_0 τ_0) ...) g))
   (where (A (@cps B C))
          (typeof ((x_0 τ_0) ... (x_i τ_i) ...) e_^’))
   (where (x_i+1 g_^’’) (next-fresh g_^’))]
  [(ANF-ctxt e Γ g) (ANF-tail e Γ g)]
  
  [(ANF-ctxt (val x : τ = e) Γ g)
   ((r_0 ... (val x : τ = e_^’)) g_^’)
   ; ANF-tail because it's OK if e_^’ is impure
   (where ((r_0 ... e_^’) g_^’) (ANF-tail e Γ g))]
  [(ANF-ctxt (def f (x : τ_1) : τ_2 = e) ((x_0 τ_0) ...) g)
   (((def f (x : τ_1) : τ_2 = (block-if-nec r_0 ... e_^’))) g_^’)
   (where ((r_0 ... e_^’) g_^’) 
          (ANF-tail e ((x τ_1) (f ((⇒ τ_1 τ_2) ε)) (x_0 τ_0) ...) g))]
  
  [(ANF-ctxt Γ g) (() g)]
  [(ANF-ctxt r_0 r_1 ... Γ g)
   ((s_0 ... s_i ...) g_^’’)
   (where ((s_0 ...) g_^’) (ANF-ctxt r_0 Γ g))
   (where ((s_i ...) g_^’’) 
          ; Don't need the inserted bindings from s_0 ...
          (ANF-ctxt r_1 ... (extend Γ (r_0)) g_^’))])

(define-metafunction SR-compile
  block-if-nec : r ... e -> e
  [(block-if-nec e) e]
  [(block-if-nec r_0 r_1 ... e)
   (block r_0 r_1 ... e)])

(define-metafunction SR-compile
  extend : Γ (r ...) -> Γ
  [(extend Γ ()) Γ]
  [(extend Γ (e r ...)) 
   (extend Γ (r ...))]
  [(extend ((x_0 τ_0) ...) ((val x : τ = e) r ...))
   (extend ((x τ) (x_0 τ_0) ...) (r ...))]
  [(extend ((x_0 τ_0) ...) ((def f (x : τ_1) : τ_2 = e) r ...))
   (extend ((f ((⇒ τ_1 τ_2) ε)) (x_0 τ_0) ...) (r ...))])

(provide ANF ANF-ctxt ANF-tail
         block-if-nec extend 
         CPS CPS’ CPS-type CPS-pure-type 
         compile)
