#lang racket

(require redex "grammar.rkt" "solve.rkt")

(provide (all-defined-out))

(define-metafunction gadt-lang
  infer : prog -> prog
  [(infer (program dd ... e_1))
   (program
     dd ...
     (finalize e_1 (infer* {} (build-delta dd ...) 0 e_1)))])

(define-metafunction gadt-lang
  finalize : e (n e F) -> e
  [(finalize e_0 (n e F))
   (finalize* e_0 e (solve F))])

(define-metafunction gadt-lang
  finalize* : e e φ -> e
  [(finalize* e_0 e ⊥) e_0]
  [(finalize* e_0 e θ) (rename θ e)]
  [(finalize* e_0 e φ) e_0])

(define-metafunction gadt-lang
  infer+ : Γ ∆ n e -> e
  [(infer+ Γ ∆ n e)
   (finalize e (infer* Γ ∆ n e))])

(define-metafunction gadt-lang
  infer* : Γ ∆ n e -> (n e F)

  ;; ERROR: no simple type case for variables
  [(infer* Γ ∆ n x)
   [n (x : τ) [^]]
   (side-condition (term (∈ x (domain Γ))))
   (where τ (lookup Γ x))]

  [(infer* Γ ∆ n x)
   [m (x : (rename φ τ)) [^]]
   (side-condition (term (∈ x (domain Γ))))
   (where (∀ (a ...) [^] ⇒ τ) (lookup Γ x))
   (where (m α ...) (fresh-α n a ...))
   (where φ {[a := α] ...})]

  [(infer* Γ ∆ n K)
   [m (K : (rename φ τ)) (rename φ C)]
   (side-condition (term (∈ K (domain ∆))))
   (where (∀ (a ...) C ⇒ τ) (lookup ∆ K))
   (where (m α ...) (fresh-α n a ...))
   (where φ {[a := α] ...})]

  [(infer* Γ ∆ n_1 (e_1 e_2))
   [n_4 (((e_3 : τ_1) (e_4 : τ_2)) : α) F]
   (where (n_2 (e_3 : τ_1) F_1) (infer* Γ ∆ n_1 e_1))
   (where (n_3 (e_4 : τ_2) F_2) (infer* Γ ∆ n_2 e_2))
   (where (n_4 α) (fresh-α n_3 (e_1 e_2)))
   (where F (& F_1 F_2 [^ (τ_1 ~ (→ τ_2 α))]))]

  [(infer* Γ ∆ n_1 (λ x e_1))
   [n_3 ((λ x (e_2 : τ)) : (→ α τ)) F]
   (where (n_2 α) (fresh-α n_1 x))
   (where (n_3 (e_2 : τ) F) (infer* (← Γ {[x := α]}) ∆ n_2 e_1))]

  ;; TYPO: `g` became `x`
  [(infer* Γ ∆ n_1 (let (x :: (∀ (a ...) [^] ⇒ τ_1) = e_1) in e_2))
   [n_3 ((let (x :: (∀ (a ...) [^] ⇒ τ_1) = (e_3 : τ_2)) in (e_4 : υ)) : υ) F]
   (where (n_2 (e_3 : τ_2) F_1)
          (infer* (← Γ {[x := (∀ (a ...) [^] ⇒ τ_1)]}) ∆ n_1 e_1))
   (where (n_3 (e_4 : υ) F_2)
          (infer* (← Γ {[x := (∀ (a ...) [^] ⇒ τ_1)]}) ∆ n_2 e_2))
   (where F (& F_2 [^ (∀ (fuv Γ) (a ...) [^] ⊃ (& F_1 [^ (τ_1 ~ τ_2)]))]))]

  ;; TYPO: `g` became `x`
  [(infer* Γ ∆ n_1 (let (x = e_1) in e_2))
   [n_5 ((let (x = (e_3* : τ)) in (e_4 : υ)) : υ) F]
   (where (n_2 α) (fresh-α n_1 x))
   (where (n_3 (e_3 : τ) F_1) (infer* (← Γ {[x := α]}) ∆ n_2 e_1))
   (where F_1* (& F_1 [^ (α ~ τ)]))
   (where F_s (simple F_1*))
   (where φ_s (solve F_s))
   (where {β ...} (∖ (fuv (rename φ_s τ)) (fuv (rename φ_s Γ))))
   (where (n_4 b ...) (fresh-τ n_3 β ...))
   (where θ_k {[β := b] ...})
   ;; MY BUG: forgot to do this renaming of e_3 at first!
   ;; Extrapolating from type judgment to full type annotation can be tricky.
   (where e_3* (rename θ_k e_3))
   (where (n_5 (e_4 : υ) F_2)
          (infer* (← Γ {[x := (∀ {b ...} [^] ⇒ (rename θ_k (rename φ_s τ)))]})
                  ∆
                  n_4
                  e_2))
   (where F (& F_2 [^ (∀ (fuv Γ) (b ...) [^] ⊃ (rename θ_k F_1*))]))]

  [(infer* Γ ∆ n_1 (case e_1 of [p → e_2] ...))
   ;; TYPO: fixed return type from α to β to get this to compile
   [n_4 ((case (e_3 : τ_e) of [p → e_4] ...) : β) F]
   (side-condition (term (has-constructor ∆ p ...)))
   (where (T a ...) (constructor ∆ p ...))
   (where (n_2 (e_3 : τ_e) F_e) (infer* Γ ∆ n_1 e_1))
   (where (n_3 β α ...) (fresh-α n_2 T a ...))
   (where (n_4 [e_4 F_p] ...)
          (constrain* n_3 Γ ∆ (→ (T α ...) β) [p → e_2] ...))
   (where F (& F_e [^ (τ_e ~ (T α ...))] F_p ...))]

  ;; default clause
  [(infer* Γ ∆ n e) [n e [^]]]

  )

(define-relation gadt-lang
  [(has-constructor ∆ (K_0 x_0 ...) (K x ...) ...)
   (∧ (∈ K_0 (domain ∆)) (∈ K (domain ∆)) ...)
   (same-cons-σ (lookup ∆ K_0) (lookup ∆ K) ...)])

(define-relation gadt-lang
  [(same-cons-σ (∀ (b ...) C ⇒ τ) ...) (same-cons-> (flatten-arrow τ) ...)])

(define-relation gadt-lang
  [(same-cons-> (→ υ ... (T a ...)) ...) (≡ (T a ...) ...)])

(define-metafunction gadt-lang
  [(constructor ∆ (K x ...) p ...)
   (T a ...)
   (where (∀ (b ...) C ⇒ τ) (lookup ∆ K))
   (where (→ υ ... (T a ...)) (flatten-arrow τ))])

(define-metafunction gadt-lang
  [(constrain* n Γ ∆ τ) (n)]
  [(constrain* n_1 Γ ∆ τ [p_0 → e_0] [p_1 → e_1] ...)
   [n_3 (e_2 F_0) (e_3 F_1) ...]
   (where [n_2 e_2 F_0] (constrain n_1 Γ ∆ τ [p_0 → e_0]))
   (where [n_3 (e_3 F_1) ...] (constrain* n_2 Γ ∆ τ [p_1 → e_1] ...))])

(define-metafunction gadt-lang
  ;; TYPO: replaced τ with α; need to record those variables
  [(constrain n_1 Γ ∆ (→ (T α ...) τ_r) [(K x_p ...) → e_1])
   [n_2 (e_2 : τ_e) F]
   (where (∀ (c ...) D ⇒ τ_k) (lookup ∆ K))
   (where [(a ... b ...) (→ υ_p ... (T a ...))]
          [(c ...) (flatten-arrow τ_k)])
   (side-condition (term (≡ {} (∩ {b ...} (∪ (ftv Γ) (ftv τ_r))))))
   ;; TYPO: replaced τ with α here too
   (where φ {[a := α] ...})
   (side-condition (term (≈ (υ_p ...) (x_p ...))))
   (where [n_2 (e_2 : τ_e) F_e]
          (infer* (← Γ (rename φ {[x_p := υ_p] ...})) ∆ n_1 e_1))
   ;; TYPO: replaced t_e with t_r, otherwise nearly everything fails!
   (where F [^ (∀ (∪ {α ...} (fuv Γ) (fuv τ_r))
                  (b ...)
                  (rename φ D)
                  ⊃
                  (& F_e [^ (τ_e ~ τ_r)]))])])

(define-metafunction gadt-lang
  fresh-α : n any ... -> (n α ...)
  [(fresh-α n) (n)]
  [(fresh-α n_1 any_1 any_2 ...)
   (n_3 α β ...)
   (where α ,(make-α (term n_1)))
   (where n_2 ,(add1 (term n_1)))
   (where (n_3 β ...) (fresh-α n_2 any_2 ...))])

(define (make-α n)
  (string->symbol
   (string-append
    (match (remainder n 3)
      [0 "α"]
      [1 "β"]
      [2 "γ"])
    (match (quotient n 3)
      [0 ""]
      [m (number->string m)]))))

(define-metafunction gadt-lang
  fresh-τ : n any ... -> (n a ...)
  [(fresh-τ n) (n)]
  [(fresh-τ n_1 any_1 any_2 ...)
   (n_3 a b ...)
   (where a ,(make-τ (term n_1)))
   (where n_2 ,(add1 (term n_1)))
   (where (n_3 b ...) (fresh-τ n_2 any_2 ...))])

(define (make-τ n)
  (string->symbol
   (string-append
    (match (remainder n 3)
      [0 "a"]
      [1 "b"]
      [2 "c"])
    ".fresh."
    (number->string (quotient n 3)))))

(define Tuple
  '[(data (Tuple a) where [MT () [^]] [More () [^] a (Tuple a)])])

(define ∆Tuple
  '{[MT := (∀ (a) [^] ⇒ (Tuple a))]
    [More := (∀ (a) [^] ⇒ (→ a (→ (Tuple a) (Tuple a))))]})

(define Unit
  '[(data (Unit) where [Make () [^]])])

(define ∆Unit
  '{[Make := (∀ () [^] ⇒ (Unit))]})

(define ∆Paper
  '{[M1 := (∀ (a) [^ (a ~ (TBool))] ⇒ (→ (TInt) (T a)))]
    [M2 := (∀ (a) [^] ⇒ (→ (TList a) (T a)))]
    [K.0 := (∀ () [^] ⇒ (TInt))]})

(define ΓPaper
  '{[x.> := (→ (TInt) (→ (TInt) (TBool)))]
    [x.null := (∀ (a) [^] ⇒ (→ (TList a) (TBool)))]})

(define ePaper
  '(λ x (case x of [(M1 y) → ((x.> y) K.0)])))

(define ePaper*
  '(λ x
      (case x of
        [(M1 y) → ((x.> y) K.0)]
        [(M2 xs) → (x.null xs)])))

(define (test-infer)
  (redex-test-suite

    (test-equal (term (infer+ ,ΓPaper ,∆Paper 0 ,ePaper)) ePaper)

    (test-equal (term (infer+ ,ΓPaper ,∆Paper 0 ,ePaper*))
      (term
        ((λ x
            ((case (x : (T γ)) of
              [(M1 y) → (((((x.> : (→ (TInt) (→ (TInt) (TBool))))
                            (y : (TInt)))
                           : (→ (TInt) (TBool)))
                          (K.0 : (TInt)))
                         : (TBool))]
              [(M2 xs) → (((x.null : (→ (TList γ) (TBool))) (xs : (TList γ)))
                          : (TBool))])
             : (TBool)))
         : (→ (T γ) (TBool)))))

    (test-equal (term (infer* {} {,@∆Tuple ,@∆Unit} 0 Make))
      (term [0 (Make : (Unit)) [^]]))
    (test-equal (term (infer (program ,@Tuple ,@Unit Make)))
      (term (program ,@Tuple ,@Unit (Make : (Unit)))))

    (test-equal (term (infer* {} {,@∆Tuple ,@∆Unit} 0 More))
      (term [1 (More : (→ α (→ (Tuple α) (Tuple α)))) [^]]))
    (test-equal (term (infer (program ,@Tuple ,@Unit More)))
      (term (program ,@Tuple
              ,@Unit
              (More : (→ α (→ (Tuple α) (Tuple α)))))))

    (test-equal (term (infer* {} {,@∆Tuple ,@∆Unit} 0 (More Make)))
      (term [2
             (((More : (→ α (→ (Tuple α) (Tuple α))))
               (Make : (Unit)))
              : β)
             [^ ((→ α (→ (Tuple α) (Tuple α))) ~ (→ (Unit) β))]]))
    (test-equal (term (infer (program ,@Tuple ,@Unit (More Make))))
      (term (program ,@Tuple
              ,@Unit
              (((More : (→ (Unit)
                         (→ (Tuple (Unit))
                          (Tuple (Unit)))))
                (Make : (Unit)))
               : (→ (Tuple (Unit)) (Tuple (Unit)))))))

    (test-equal (term (infer (program ,@Tuple ,@Unit ((More Make) MT))))
      (term (program ,@Tuple
              ,@Unit
              (((((More : (→ (Unit)
                           (→ (Tuple (Unit))
                            (Tuple (Unit)))))
                  (Make : (Unit)))
                 : (→ (Tuple (Unit)) (Tuple (Unit))))
                (MT : (Tuple (Unit))))
               : (Tuple (Unit))))))

    (test-equal (term (infer (program (λ x x))))
      (term (program ((λ x (x : α)) : (→ α α)))))

    (test-equal (term (infer (program ,@Unit ((λ x x) Make))))
      (term (program ,@Unit
              ((((λ x (x : (Unit))) : (→ (Unit) (Unit)))
                (Make : (Unit)))
               : (Unit)))))

    (test-equal (term (infer (program (let (x :: (∀ (a) [^] ⇒ a) = x)
                                        in
                                        x))))
      (term (program ((let (x :: (∀ (a) [^] ⇒ a)
                              = (x : a))
                        in
                        (x : β))
                      : β))))

    (test-equal (term (infer (program (let (x :: (∀ (a) [^] ⇒ (→ a a)) = (λ y y))
                                        in
                                        x))))
      (term (program ((let (x :: (∀ (a) [^] ⇒ (→ a a))
                              = ((λ y (y : a)) : (→ a a)))
                        in
                        (x : (→ β β)))
                      : (→ β β)))))

    (test-equal (term (infer (program ,@Tuple
                               (let (x :: (∀ (a) [^] ⇒ (Tuple a)) = MT)
                                 in
                                 x))))
      (term (program ,@Tuple
              ((let (x :: (∀ (a) [^] ⇒ (Tuple a))
                       = (MT : (Tuple a)))
                 in
                 (x : (Tuple β)))
               : (Tuple β)))))

    (test-equal (term (infer (program ,@Tuple
                               ,@Unit
                               (let (x :: (∀ (a) [^] ⇒ (→ a (Tuple a)))
                                       = (λ y MT))
                                 in
                                 (x Make)))))
      (term (program ,@Tuple
              ,@Unit
              ((let (x :: (∀ (a) [^] ⇒ (→ a (Tuple a)))
                       = ((λ y (MT : (Tuple a)))
                          : (→ a (Tuple a))))
                 in
                 (((x : (→ (Unit) (Tuple (Unit))))
                   (Make : (Unit)))
                  : (Tuple (Unit))))
               : (Tuple (Unit))))))

    (test-equal (term (infer (program (let (x = x) in x))))
      (term (program ((let (x = (x : α))
                        in
                        (x : γ)) : γ))))

    (test-equal (term (infer (program ,@Tuple
                               ,@Unit
                               (let (x = Make) in (More x)))))
      (term (program ,@Tuple
              ,@Unit
              ((let (x = (Make : (Unit)))
                 in
                 (((More : (→ (Unit)
                            (→ (Tuple (Unit))
                             (Tuple (Unit)))))
                   (x : (Unit)))
                  : (→ (Tuple (Unit)) (Tuple (Unit)))))
               : (→ (Tuple (Unit)) (Tuple (Unit)))))))

    (test-equal (term (infer (program ,@Unit (case Make of [(Make) → Make]))))
      (term (program ,@Unit
              ((case (Make : (Unit))
                 of
                 [(Make) → (Make : (Unit))])
               : (Unit)))))

    (test-equal
      (term
        (infer
          (program
            ,@Tuple
            ,@Unit
            (let
              (x = (λ y
                      (case y
                        of
                        [(MT) → Make]
                        [(More z1 z2) → (x z2)])))
              in
              (x ((More Make) MT))))))
      (term
        (program
          ,@Tuple
          ,@Unit
          ((let
             (x = ((λ y
                      ((case (y : (Tuple c.fresh.1))
                         of
                         [(MT) → (Make : (Unit))]
                         [(More z1 z2)
                          →
                          (((x : (→ (Tuple c.fresh.1) (Unit)))
                            (z2 : (Tuple c.fresh.1)))
                           : (Unit))])
                       : (Unit)))
                   : (→ (Tuple c.fresh.1) (Unit))))
             in
             (((x : (→ (Tuple (Unit)) (Unit)))
               (((((More : (→ (Unit)
                            (→ (Tuple (Unit))
                             (Tuple (Unit)))))
                   (Make : (Unit)))
                  : (→ (Tuple (Unit)) (Tuple (Unit))))
                 (MT : (Tuple (Unit))))
                : (Tuple (Unit))))
              : (Unit)))
           : (Unit)))))))
