#lang racket

(require redex "grammar.rkt" "helper.rkt")

(provide (all-defined-out))

(define-metafunction gadt-lang
  solve : F -> φ
  [(solve F)
   (compose θ φ)
   (where F_s (simple F))
   (where φ (solve* F_s))
   (where θ (solve* (rename φ F)))])

(define-metafunction gadt-lang
  solve* : F -> φ
  [(solve* [^]) {}]
  [(solve* [^ FF_1 FF_2 FF_3 ...])
   (compose φ_2 φ_1)
   (where φ_1 (solve* [^ FF_2 FF_3 ...]))
   (where φ_2 (solve* (rename φ_1 [^ FF_1])))]
  [(solve* [^ (τ ~ τ)]) {}]
  ;; ERROR: arrow types not included in solution
  [(solve* [^ ((→ τ_1 υ_1) ~ (→ τ_2 υ_2))])
   (solve* [^ (τ_1 ~ τ_2) (υ_1 ~ υ_2)])]
  [(solve* [^ ((T τ ...) ~ (T υ ...))])
   (solve* [^ (τ ~ υ) ...])
   (side-condition (term (≈ (τ ...) (υ ...))))]
  ;; ERROR : did not prefer to assign α's
  [(solve* [^ (α ~ τ)])
   {[α := τ]}
   (side-condition (term (¬ (references α τ))))]
  [(solve* [^ (τ ~ α)])
   {[α := τ]}
   (side-condition (term (¬ (references α τ))))]
  [(solve* [^ (ν ~ τ)])
   {[ν := τ]}
   (side-condition (term (¬ (references ν τ))))]
  [(solve* [^ (τ ~ ν)])
   {[ν := τ]}
   (side-condition (term (¬ (references ν τ))))]
  [(solve* [^ (∀ [α ...] (b ...) [^] ⊃ F)])
   φ
   (where φ (solve* F))
   (side-condition (term (≡ {} (∩ {b ...} (∪ (fv (rename φ α)) ...)))))
   (side-condition (term (≡ {} (∩ {b ...} (domain φ)))))]
  [(solve* [^ (∀ [α ...] (b ...) C ⊃ F)])
   θ
   (where φ (solve* C))
   (where θ (solve (rename φ F)))
   (side-condition (term (≡ {} (∩ {α ...} (domain θ)))))]
  [(solve* F) ⊥])

(define-relation gadt-lang
  [(references ν ν)]
  [(references ν (T τ ...)) (∨ (references ν τ) ...)]
  [(references ν (→ τ υ)) (∨ (references ν τ) (references ν υ))])

(define-metafunction gadt-lang
  simple : F -> F
  [(simple C) C]
  [(simple [^ (∀ [α ...] (b ...) [^] ⊃ F)])
   [^ (∀ [α ...] (b ...) [^] ⊃ (simple F))]]
  [(simple [^ FF]) [^]]
  [(simple [^ FF ...]) (& (simple [^ FF]) ...)])

(define-metafunction gadt-lang
  compose : φ ... -> φ
  [(compose) {}]
  [(compose φ) φ]
  [(compose φ_1 ... φ_2)
   (⊎ φ_3 (rename φ_3 φ_2))
   (where φ_3 (compose φ_1 ...))])

(define (test-solve)
  (redex-test-suite
    (test-reorder (term (solve [^])) (term {}))
    (test-reorder (term (solve [^ (α ~ (T))])) (term {[α := (T)]}))
    (test-reorder (term (solve [^ ((T) ~ α)])) (term {[α := (T)]}))
    (test-reorder (term (solve [^ ((T) ~ α) (α ~ (S))])) (term ⊥))
    (test-reorder (term (solve [^ (α ~ (→ β (T))) (α ~ (→ (S) γ))]))
      (term {[α := (→ (S) (T))] [β := (S)] [γ := (T)]}))))
